{-# LANGUAGE NamedFieldPuns #-}
import Data.List
import E
import ES
import Term
import Substitution
import PE
import PES
import TPTP
import UEQParser
import INFParser
import TRSParser
import Horn
import Prover
import Util
import ReductionOrder
import Termination
import System.Environment

-- INF: CoCo infeasibility problem
-- Termination: termination checker
data Mode = UEQ String
          | TPTP String
          | INF String
          | Help
          | Waldmeister String
          | OrderedCompletion String
          | Termination String (ReductionOrder.Class)

-- NOTE: default order should be GWPO01?
defaultConfig :: Prover.Config
defaultConfig = Prover.Config { inter_reduction = True, order_class = ReductionOrder.LPO, verbose = False }

convertEquation :: PE -> TPTPInput
convertEquation (l, r) = TPTP.CNF "some_equation" "axiom" [(True, Equation l r)]

convertGoal :: PE -> TPTPInput
convertGoal (l, r) = TPTP.CNF "some_equation" "negated_conjecture" [(False, Equation l r)]

convertES :: PES -> TPTP
convertES es = [ convertEquation e | e <- es ]

-- TODO: hacky implementation
parseArgs' :: [String] -> Mode -> Prover.Config -> (Mode, Prover.Config)
parseArgs' [] m c = (m, c)
-- help
parseArgs' ("-h" : _) _ _ =  (Help, defaultConfig)
parseArgs' ("--help" : _) _ _ = (Help, defaultConfig)
-- termination
parseArgs' ("--termination" : fname : []) _ _ = (Termination fname ReductionOrder.GWPO, defaultConfig)
parseArgs' ("--termination" : fname : "--lpo" : []) _ _ = (Termination fname ReductionOrder.LPO, defaultConfig)
parseArgs' ("--termination" : fname : "--gwpo" : []) _ _ = (Termination fname ReductionOrder.GWPO, defaultConfig)
parseArgs' ("--termination" : fname : "--gwpoN" : []) _ _ = (Termination fname ReductionOrder.GWPON, defaultConfig)
parseArgs' ("--termination" : fname : "--gwpo01" : []) _ _ = (Termination fname ReductionOrder.GWPO01, defaultConfig)
parseArgs' ("--termination" : fname : "--wpo" : []) _ _ = (Termination fname ReductionOrder.WPO, defaultConfig)
parseArgs' ("--termination" : fname : "--lpo-via-gwpo" : []) _ _ = (Termination fname ReductionOrder.LPOviaGWPO, defaultConfig)
-- prover
parseArgs' ("--ueq" : fname : args) _ c = parseArgs' args (UEQ fname) c
parseArgs' ("--tptp" : fname : args) _ c = parseArgs' args (TPTP fname) c
parseArgs' ("--inf" : fname : args) _ c = parseArgs' args (INF fname) c
parseArgs' ("--waldmeister" : fname : args) _ c = parseArgs' args (Waldmeister fname) c
-- ordered completion
parseArgs' ("--ordered-completion" : fname : args) _ c = parseArgs' args (OrderedCompletion fname) c
-- options
parseArgs' ("--no-inter-reduction" : args) m c = parseArgs' args m (Config { inter_reduction = False, order_class = order_class c, verbose = verbose c })
parseArgs' ("--no-simplification" : args) m c = parseArgs' args m (Config { inter_reduction = False, order_class = order_class c, verbose = verbose c })
parseArgs' ("--lpo" : args) m c = parseArgs' args m (Config { inter_reduction = inter_reduction c, order_class = LPO, verbose = verbose c})
parseArgs' ("--gwpo" : args) m c = parseArgs' args m (Config { inter_reduction = inter_reduction c, order_class = GWPO, verbose = verbose c})
parseArgs' ("--gwpo01" : args) m c = parseArgs' args m (Config { inter_reduction = inter_reduction c, order_class = GWPO01, verbose = verbose c})
parseArgs' ("--gwpoN" : args) m c = parseArgs' args m (Config { inter_reduction = inter_reduction c, order_class = GWPON, verbose = verbose c})
parseArgs' ("--wpo" : args) m c = parseArgs' args m (Config { inter_reduction = inter_reduction c, order_class = WPO, verbose = verbose c})
parseArgs' ("--lpo-via-gwpo" : args) m c = parseArgs' args m (Config { inter_reduction = inter_reduction c, order_class = LPOviaGWPO, verbose = verbose c})
parseArgs' ("--verbose" : args) m c = parseArgs' args m (Config { inter_reduction = inter_reduction c, order_class = order_class c, verbose = True})
parseArgs' ("--v" : args) m c = parseArgs' args m (Config { inter_reduction = inter_reduction c, order_class = order_class c, verbose = True})
parseArgs' (fname : args) _ c = parseArgs' args (UEQ fname) c

parseArgs :: [String] -> (Mode, Prover.Config)
parseArgs args = parseArgs' args Help defaultConfig

version :: String
version = "0.3"

handleHelp :: IO ()
handleHelp = do
  putStrLn ("toma version " ++ version)
  putStrLn "Usage: toma <options> <file>"
  putStrLn "<file> is in TPTP (UEQ) format, CoCo INF format or TRS format."
  putStrLn "equational theorem proving: "
  putStrLn "$ toma inverse_unit.p        # TPTP UEQ format (simple skolemization)"
  putStrLn "$ toma --ueq  inverse_unit.p # TPTP UEQ format (simple skolemization)"
  putStrLn "$ toma --waldmeister inverse_unit.p # TPTP UEQ format (Waldmeister transformation)"
  putStrLn "$ toma --tptp inverse_unit.p # TPTP format (split-if transformation)"
  putStrLn "$ toma --inf  936.trs        # CoCo INF format (split-if transformation)"
  putStrLn "debugging prover:"
  putStrLn "$ toma --ueq inverse_unit.p --verbose  # verbose output for equational reasoning"
  putStrLn "$ toma --ueq inverse_unit.p --v        # the same as --verbose" 
  putStrLn "equational theorem proving with various reduciton orders"
  putStrLn "no simplification (or no inter-reduction):"
  putStrLn "$ toma --ueq  inverse_unit.p --no-simplification"
  putStrLn "$ toma --ueq  inverse_unit.p --no-inter-reduction"
  putStrLn "default ordering is --gwpo01"
  putStrLn "$ toma inverse_unit.p                # the same as --gwpo01"
  putStrLn "$ toma --lpo inverse_unit.p          # LPO"
  putStrLn "$ toma --gwpo inverse_unit.p         # the same as --gwpo01"
  putStrLn "$ toma --gwpoN inverse_unit.p        # GWPO with linear polynomial interpretation over N"
  putStrLn "$ toma --gwpo01 inverse_unit.p       # GWPO with linear polynomial interpretation over {0, 1} for coefficients and N for constants"
  putStrLn "$ toma --wpo inverse_unit.p          # WPO with linear polynomial interpretation over N"
  putStrLn "$ toma --lpo-via-gwpo inverse_unit.p # LPO simulation via GWPO"
  putStrLn "ordered maximal completion with various reduction orders: "
  putStrLn "$ toma --ordered-completion group.trs          # LPO "
  putStrLn "$ toma --ordered-completion group.trs --lpo    # LPO"
  putStrLn "$ toma --ordered-completion group.trs --gwpo01 # GWPO with linear polynomial interpretation over {0, 1} for coefficients and N for constants"
  putStrLn "termination checking with reduction orders: "
  putStrLn "$ toma --termination ack.trs                 # the same as --gwpo"
  putStrLn "$ toma --termination ack.trs --lpo           # LPO"
  putStrLn "$ toma --termination ack.trs --gwpoN         # GWPO with linear polynomial interpretation over N"
  putStrLn "$ toma --termination ack.trs --gwpo01        # GWPO with linear polynomial interpretation over {0, 1} for coefficients and N for constants"
  putStrLn "$ toma --termination ack.trs --gwpo          # first tries --gwpo01, and then --gwpoN"
  putStrLn "$ toma --termination ack.trs --wpo           # WPO with linear polynomial interpretation over N"
  putStrLn "$ toma --termination ack.trs --lpo-via-gwpo  # LPO simulation via GWPO"

printTPTPProof :: PES -> Proof -> IO ()
printTPTPProof pes (Prover.Join { proof_goal, proof_es, proof_deleted_es }) = do
  putStrLn "% SZS status Unsatisfiable"
  putStrLn "% SZS output start Proof"
  putStrLn "The given problem is transformed into a word problem"
  putStrLn "by the split-if encoding (Claessen and Smallbone, 2018)."
  putStrLn "To show the unsatisfiability,"
  putStrLn "it suffices to show that true__ = false__ is valid in the following ES:"
  putStrLn (showPES pes)
  putStrLn "This is an equational proof of true__ = false__:"
  putStrLn (showES (sortById (relevant (eqn_id proof_goal) (proof_goal : proof_es ++ proof_deleted_es))))
  putStrLn "% SZS output end Proof"
printTPTPProof pes (Prover.Complete { proof_es, proof_reduction_order_param }) = do
  putStrLn "% SZS status Satisfiable"
  putStrLn "% SZS output start Proof"
  putStrLn "The given problem is transformed into a word problem"
  putStrLn "by the split-if encoding (Claessen and Smallbone, 2018)."
  putStrLn "To show the satisfiability,"
  putStrLn "it suffices to show that true__ = false__ is not valid in the following ES:"
  putStrLn (showPES pes)
  putStrLn "The ES admits the following ground-complete ordered rewrite system:"
  putStrLn (show proof_reduction_order_param)
  putStrLn "ES:"
  putStrLn (showOTRS (param2ord proof_reduction_order_param) proof_es)
  putStrLn "Since true__ and false__ are not joinable in the ordered rewrite system,"
  putStrLn " true__ = false__ is not valid."
  putStrLn "% SZS output end Proof"
printTPTPProof _ Failure = putStrLn "% SZS status GaveUp : proof failed"

printTPTPTrivial :: IO ()
printTPTPTrivial = do
  putStrLn "% SZS status Satisfiable"
  putStrLn "% SZS output Proof"
  putStrLn "% The given problem admits an trivial model."
  putStrLn "% SZS output end Proof"

-- TODO: suppert 'include' from $TPTP 
handleTPTP :: String -> Prover.Config -> IO ()
handleTPTP fname conf = do
  result <- readTPTP fname
  case result of
    Left e -> do
      putStdErr "% SZS status SyntaxError"
      putStdErr (show e)
    Right tptp -> do
      case splitIfTPTP tptp of
        Nothing      -> do
          putStdErr "% SZS status UsageError : The input contains a non-Horn-clause."
        Just (es, goal@(s, t)) -> do
          putStr (showTPTP (convertES es ++ [convertGoal goal]))
          if s == t -- trivial model found
            then printTPTPTrivial
            else do p <- prove conf goal (axioms es)
                    printTPTPProof es p

separate :: [AnnotatedFormula] -> (PES, PES)
separate [] = ([], [])
separate ((_, _, (Eq, pe)) : afs) =
  (pe : pes, gs)
  where (pes, gs) = separate afs
separate ((_, _, (Neq, g)) : afs) = 
  (pes, g : gs)
  where (pes, gs) = separate afs

skolemize :: [String] -> (Term, Term) -> (Term, Term)
skolemize fs (s, t) =
  (substitute s sigma, substitute t sigma)
  where
    xs = Data.List.union (variables s) (variables t)
    cs = [ c | i <- [1 :: Int ..], let c = "c" ++ show i, notElem c fs ]
    sigma = [ (x, F c []) | (x, c) <- zip xs cs ]

syms :: String -> [String]
syms s = s : [s ++ show i | i <- [0 :: Int .. ] ]

-- transforms a (possiblely non-ground) goal s = t
-- into __true = __false
-- by adding __eq(x, x) = __true
--           __eq(s, t) = __false
waldmeister :: PES -> PE -> (PES, PE)
waldmeister pes goal@(s, t) = (eqAxiom : gAxiom : pes, (tTerm, fTerm))
  where
    fs = functionsInES (goal : pes) 
    eqSymbol = head [ c | c <- syms "__eq" , notElem c fs]
    tSymbol = head [ c | c <- syms "__true" , notElem c fs]
    fSymbol = head [ c | c <- syms "__false" , notElem c fs]
    tTerm = F tSymbol []
    fTerm = F fSymbol []
    eqAxiom = (F eqSymbol [V "X", V "X"], tTerm)
    gAxiom = (F eqSymbol [s, t], fTerm)

-- add fake goal for ordered completion
fakeGoal :: PES -> PE
fakeGoal pes = (tTerm, fTerm)
  where
    fs = functionsInES pes
    tSymbol = head [ c | c <- syms "__true" , notElem c fs]
    fSymbol = head [ c | c <- syms "__false" , notElem c fs]
    tTerm = F tSymbol []
    fTerm = F fSymbol []

includeDirectories :: IO [String]
includeDirectories = do
  env <- getEnvironment
  case lookup "TPTP" env of
    Nothing  -> return ["."]
    Just dir -> return [".", dir]

printUEQProof :: PE -> PE -> PES -> Proof -> IO ()
printUEQProof _ sko_goal _pes (Prover.Join { proof_goal, proof_es, proof_deleted_es}) = do
  putStrLn "% SZS status Unsatisfiable"
  putStrLn "% SZS output start Proof"
  putStrLn "To show the unsatisfiability,"
  putStrLn ("it suffices to show that " ++ showPE sko_goal   ++ " (skolemized goal) is valid under the axioms.")
  putStrLn "This is an equational proof:"
  putStrLn (showES (sortById (relevant (eqn_id proof_goal) (proof_goal : proof_es ++ proof_deleted_es))))
  putStrLn "% SZS output end Proof"
printUEQProof orig_goal@(s, t) _ pes (Prover.Complete { proof_es, proof_reduction_order_param })
  | PE.isGround orig_goal = do
      putStrLn "% SZS status Satisfiable"
      putStrLn "% SZS output start Proof"
      putStrLn "To show the satisfiability,"
      putStrLn ("it suffices to show that " ++ showPE orig_goal ++ " is not valid under the axioms:")
      putStrLn (showPES pes)
      putStrLn "The ES admits the following ground-complete ordered rewrite system:"
      putStrLn (show proof_reduction_order_param)
      putStrLn "ES:"
      putStrLn (showOTRS (param2ord proof_reduction_order_param) proof_es)
      putStrLn ("Since " ++ show s ++ " and " ++ show t ++ " are not joinable in the ordered rewrite system,")
      putStrLn ("the ground goal " ++ showPE orig_goal ++ " is not valid.")
      putStrLn "% SZS output end Proof"
  | otherwise = do
      putStrLn "% SZS status GaveUp : goal is not ground"
      putStrLn "The ES admits the following ground-complete ordered rewrite system:"
      putStrLn (show proof_reduction_order_param)
      putStrLn "ES:"
      putStrLn (showOTRS (param2ord proof_reduction_order_param) proof_es)
printUEQProof _ _ _ Failure = putStrLn "% SZS status GaveUp : proof failed"

-- pure equational theory is always consistent in the sense of first-order logic
-- since it has trivial model (model with one element).
-- Namely it is satisfiable in the sense of first-order logic.
handleNoGoal :: IO ()
handleNoGoal = do
  putStrLn "% SZS status Satisfiable"
  putStrLn "% SZS output start Proof"
  putStrLn "The theory has the trivial model."
  putStrLn "% SZS output end Proof"

handleTooManyGoals :: IO ()
handleTooManyGoals = putStdErr "% SZS status UsageError : number of goals must be exactly one"

handleUEQ :: String -> Prover.Config -> IO ()
handleUEQ filename conf = do
  dirs <- includeDirectories
  result <- readUEQFile dirs filename
  case result of
    Left e -> do
      putStdErr "% SZS status SyntaxError"
      putStdErr (show e)
    Right afs -> do
      let (pes, goals) = separate afs
      case goals of
        [goal] -> do
          let fs = functionsInES (goal : pes) 
          let skolemized_goal = skolemize fs goal
          let es = axioms pes
          p <- prove conf skolemized_goal es
          printUEQProof goal skolemized_goal (toPES es) p
        [] -> handleNoGoal
        _ -> handleTooManyGoals

printWaldmeisterProof :: PE -> PE -> PES -> Proof -> IO ()
printWaldmeisterProof _ new_goal _pes (Prover.Join { proof_goal, proof_es, proof_deleted_es}) = do
  putStrLn "% SZS status Unsatisfiable"
  putStrLn "% SZS output start Proof"
  putStrLn "To show the unsatisfiability,"
  putStrLn ("it suffices to show that " ++ showPE new_goal   ++ " (skolemized goal) is valid under the axioms.")
  putStrLn "This is an equational proof:"
  putStrLn (showES (sortById (relevant (eqn_id proof_goal) (proof_goal : proof_es ++ proof_deleted_es))))
  putStrLn "% SZS output end Proof"
printWaldmeisterProof orig_goal new_goal@(s, t) pes (Prover.Complete { proof_es, proof_reduction_order_param }) = do
  putStrLn "% SZS status Satisfiable"
  putStrLn "% SZS output start Proof"
  putStrLn ("To show the satisfiability of " ++ showPE orig_goal ++ ",")
  putStrLn ("it suffices to show that " ++ showPE new_goal ++ " is not valid under the axioms:")
  putStrLn (showPES pes)
  putStrLn "The ES admits the following ground-complete ordered rewrite system:"
  putStrLn (show proof_reduction_order_param)
  putStrLn "ES:"
  putStrLn (showOTRS (param2ord proof_reduction_order_param) proof_es)
  putStrLn ("Since " ++ show s ++ " and " ++ show t ++ " are not joinable in the ordered rewrite system,")
  putStrLn ("the original goal " ++ showPE orig_goal  ++ " is not valid.")
  putStrLn "% SZS output end Proof"
printWaldmeisterProof _ _ _ Failure = putStrLn "% SZS status GaveUp : proof failed"

-- same as handleUEQ, exept for transformation waldmeister
handleWaldmeister :: String -> Prover.Config -> IO ()
handleWaldmeister filename conf = do
  dirs <- includeDirectories
  result <- readUEQFile dirs filename
  case result of
    Left e -> do
      putStdErr "% SZS status SyntaxError"
      putStdErr (show e)
    Right afs -> do
      let (pes, goals) = separate afs
      case goals of
        [goal] -> do
          let (pes', goal') = waldmeister pes goal
          let axs = axioms pes'
          p <- prove conf goal' axs
          printWaldmeisterProof goal goal' (toPES axs) p
        [] -> handleNoGoal
        _ -> handleTooManyGoals

printINFProof :: PES -> Proof -> IO ()
printINFProof pes (Complete {proof_es, proof_reduction_order_param} ) = do
  putStrLn "YES"
  putStrLn "Proof: "
  putStrLn "To show the infeasibility of the given CoCo INF problem, "
  putStrLn "we transform the problem into a word problem by the split-if encoding (Claessen and Smallbone, 2018)."
  putStrLn "It suffices to show that the follwing ES does not entail true__ = false__ ."
  putStrLn "ES:"
  putStrLn (showPES pes)
  putStrLn "The ES admits the following equivalent ground-complete ordered rewrite system:"
  putStrLn (show proof_reduction_order_param)
  putStrLn "ES:"
  putStrLn (showOTRS (param2ord proof_reduction_order_param) proof_es)
  putStrLn "Since true__ and false__ are not joinable,"
  putStrLn "the original problem is infeasible."
printINFProof _ (Prover.Join {}) = putStrLn "MAYBE"
printINFProof _ Failure = putStrLn "MAYBE"

handleINF :: String -> Prover.Config -> IO ()
handleINF fname conf = do
  result <- readINFProblem fname
  case result of
    Left e -> do
      putStdErr "ERROR"
      putStdErr (show e)
    Right inf -> do
      case splitIfINF inf of
        Nothing      -> putStdErr "ERROR"
        Just (es, goal@(s, t)) -> do
          if s == t -- trivial model found
            then do putStrLn "MAYBE" -- TODO: print YES with an horn clauses that admits a trivial model.
                    -- putStrLn "Proof:"
                    -- putStrLn "The given CoCo INF problem is transformed into a satisfiability problem of horn clauses."
                    -- -- TODO: we need to print the horn clauses here...
                    -- putStrLn "Since the horn clauses has a trivial model,"
                    -- putStrLn "the original problem is infeasible."
            else do p <- prove conf goal (axioms es)
                    printINFProof es p

printOrderedCompletionProof :: PE -> Proof -> IO ()
printOrderedCompletionProof fake_goal (Prover.Join {}) = do
  putStrLn "ERROR"
  putStrLn ("The joinability of the fake goal " ++ (showPE fake_goal) ++ " implies the existence of a bug.")
printOrderedCompletionProof fake_goal (Prover.Complete { proof_es, proof_reduction_order_param }) = do
  putStrLn "YES"
  putStrLn "The ES admits the following ground-complete ordered rewrite system:"
  putStrLn (show proof_reduction_order_param)
  putStrLn "ES:"
  putStrLn (showOTRS (param2ord proof_reduction_order_param) proof_es)
  putStrLn ""
  putStrLn ("Note that new symbols are introduced by the fake goal " ++ (showPE fake_goal) ++ ".")
printOrderedCompletionProof _ Failure = do
  putStrLn "ERROR"
  putStrLn "completion failed"

handleOrderedCompletion :: String -> Prover.Config -> IO ()
handleOrderedCompletion fname conf = do
  result <- readTRSFile fname
  case result of
    Left e -> do
      putStdErr "ERROR"
      putStdErr (show e)
    Right (pes, _mu) -> do
      let axs = axioms pes
      let fake_goal = fakeGoal pes
      p <- prove conf fake_goal axs
      printOrderedCompletionProof fake_goal p

handleTermination :: String -> ReductionOrder.Class -> IO ()
handleTermination fname conf = do
  parseres <- readTRSFile fname
  case parseres of
    Left e -> do
      putStdErr "ERROR"
      putStdErr (show e)
    Right (trs, _mu) ->
      terminationChecker conf trs

dispatch :: (Mode, Prover.Config) -> IO ()
dispatch (Help, _conf) = handleHelp
dispatch (UEQ fname, conf) = handleUEQ fname conf
dispatch (Waldmeister fname, conf) = handleWaldmeister fname conf
dispatch (TPTP fname, conf) = handleTPTP fname conf
dispatch (INF fname, conf) = handleINF fname conf
dispatch (OrderedCompletion fname, conf) = handleOrderedCompletion fname conf
dispatch (Termination fname tconf, _conf) = handleTermination fname tconf

main :: IO ()
main = do
  args <- getArgs
  dispatch (parseArgs args)
