module Prover where

import Data.List
import E
import ES
import Derivation
import PE
import PES
import SMT
import Z3
import Term
import LPOEncoding
import Rewriting
import CP
import KBCompletion
import Substitution
import ReductionOrder
import Util
import GWPO

sortByWeight :: [(PE, Derivation)] -> [(PE, Derivation)]
sortByWeight cs = sortBy comp cs
  where comp (e1, _) (e2, _) = compare (PE.size e1) (PE.size e2)

selectLemmata :: [(PE, Derivation)] -> [(PE, Derivation)]
selectLemmata cs = take n (Prover.sortByWeight cs)
  where n = 12

-- NOTE: The prefix "proof_" is to avoid shadow-binding warnings.
data Proof = Join { proof_goal :: E, proof_es :: ES, proof_deleted_es :: ES }
           -- system is ground-complete but goal is not joinable.
           | Complete { proof_es :: ES, proof_reduction_order_param :: ReductionOrderParam , proof_deleted_es :: ES}
           | Failure -- not used for now

findLPOPrec :: PES -> PE -> IO Precedence
findLPOPrec es goal = do
  res <- z3 "toma" z3Input
  case res of
    Nothing -> error "findLPOPrec: no precedence found." -- this error means bug?
    Just res' -> return (assoc2Prec [ (fname, n) | (fname, v) <- res', n <- fromSMTValue2Int v])
  where
    esWithIds = [("__rulevar" ++ show i, e) | (i, e) <- zip ([0..] :: [Int]) (es ++ PES.inverse es)]
    -- NOTE: include goal (reduction orders require total precedence)
    functionSymbols = nub [f | (l, r) <- goal : es, f <- (Term.functions l) ++ (Term.functions r)]
    esIds = [ esId | (esId, _) <- esWithIds ]
    reducibilityOnTerm t =
      orFormula [ BoolVar ruleId | (ruleId, (l, _)) <- esWithIds, encompass t l ]
    test_terms = nub [ u | (s, t) <- goal : es, u <- [s, t] ]
    maximizedConstraints = [ reducibilityOnTerm t | t <- test_terms ]
    lpoDic = constructDic (map snd esWithIds)
    dicCostraints = andFormula [ impliesFormula (BoolVar name) (encodeLPODic lpoDic s t) | ((s, t), name) <- lpoDic]
    orientation = andFormula [ impliesFormula (BoolVar ruleId) (BoolVar (unsafeLookup (l, r) lpoDic))
                             | (ruleId, (l, r)) <- esWithIds ]
    -- totality is required since precedence is presented as list
    precTotality = Distinct [Var f | f <- functionSymbols]
    z3Input = maxsat functionSymbols
                     (esIds ++ map snd lpoDic)
                     maximizedConstraints
                     (andFormula [orientation, precTotality, dicCostraints])

findGWPOParamN :: PES -> PE -> IO WPOParam
findGWPOParamN es goal = do
  res <- z3 "toma" z3Input
  case res of
    Nothing -> error "findGWPOParamN: no parameter is found." -- this error means bug?
    Just res' ->
      let prec = assoc2Prec [ (f, n) | f <- z3Vars_prec, SMTIntValue n <- [unsafeLookup f res'] ]
          alg = [ (c, smtValueUnsafeCastInt (unsafeLookup c res')) | c <- z3Vars_coef ]
      in return (alg, prec)
  where
    esWithIds = [("__rulevar" ++ show i, e) | (i, e) <- zip ([0..] :: [Int]) (es ++ PES.inverse es)]
    -- NOTE: include goal (reduction orders require total precedence)
    fs = functionsInESWithArity (goal : es)
    reducibilityOnTerm t =
      orFormula [ BoolVar ruleId | (ruleId, (l, _)) <- esWithIds, encompass t l ]
    test_terms = nub [ u | (s, t) <- goal : es, u <- [s, t] ]
    maximizedConstraints = [ reducibilityOnTerm t | t <- test_terms ]
    -- constraits to be satisfied
    orientation = andFormula [ impliesFormula (BoolVar ruleId) (encodeGWPO N l r)
                             | (ruleId, (l, r)) <- esWithIds ]
    precTotality = Distinct [Var f | f <- z3Vars_prec ] -- totality is required since precedence is presented as list
    coef_nonneg = andFormula [ Geq (Var v) (Val 0) | v <- z3Vars_coef ] -- weak monotonicity
    -- variables
    z3Vars_coef = [ ith_coef f i | (f, n) <- fs, i <- [0..n] ]
    z3Vars_prec = [ prec_name f | (f, _n) <- fs ]
    esIds = [ esId | (esId, _) <- esWithIds ]
    z3Input = maxsat (z3Vars_prec ++ z3Vars_coef)
                     esIds
                     maximizedConstraints
                     (andFormula [orientation, precTotality, coef_nonneg])

findGWPOParam01 :: PES -> PE -> IO WPOParam
findGWPOParam01 es goal = do
  res <- z3 "toma" z3Input
  case res of
    Nothing -> error "findGWPOParam01: no parameter is found." -- this error means bug?
    Just res' ->
      let prec = assoc2Prec [ (f, n) | f <- z3Vars_prec, SMTIntValue n <- [unsafeLookup f res'] ]
          alg = [ (c, smtValueUnsafeCastInt (unsafeLookup c res')) | c <- z3Vars_coef1n ++ z3Vars_coef0 ]
      in return (alg, prec)
  where
    esWithIds = [("__rulevar" ++ show i, e) | (i, e) <- zip ([0..] :: [Int]) (es ++ PES.inverse es)]
    -- NOTE: include goal (reduction orders require total precedence)
    fs = functionsInESWithArity (goal : es)
    reducibilityOnTerm t =
      orFormula [ BoolVar ruleId | (ruleId, (l, _)) <- esWithIds, encompass t l ]
    test_terms = nub [ u | (s, t) <- goal : es, u <- [s, t] ]
    maximizedConstraints = [ reducibilityOnTerm t | t <- test_terms ]
    -- constraits to be satisfied
    orientation = andFormula [ impliesFormula (BoolVar ruleId) (encodeGWPO ZeroOne l r)
                             | (ruleId, (l, r)) <- esWithIds ]
    precTotality = Distinct [Var f | f <- z3Vars_prec ] -- totality is required since precedence is presented as list
    coef0_nonneg = andFormula [ Geq (Var v) (Val 0) | v <- z3Vars_coef0 ] -- this may not be neccesary.
    -- variables
    z3Vars_coef1n = [ ith_coef f i | (f, n) <- fs, i <- [1..n] ]
    z3Vars_coef0 = [ ith_coef f 0 | (f, _n) <- fs ]
    z3Vars_prec = [ prec_name f | (f, _n) <- fs ]
    esIds = [ esId | (esId, _) <- esWithIds ]
    z3Input = maxsat (z3Vars_prec ++ z3Vars_coef0)
                     (esIds ++ z3Vars_coef1n)
                     maximizedConstraints
                     (andFormula [orientation, precTotality, coef0_nonneg])


findWPOParam :: PES -> PE -> IO WPOParam
findWPOParam es goal = do
  res <- z3 "toma" z3Input
  case res of
    Nothing -> error "findWPOParam: no parameter is found." -- this error means bug?
    Just res' ->
      let prec = assoc2Prec [ (f, n) | f <- z3Vars_prec, SMTIntValue n <- [unsafeLookup f res'] ]
          alg = [ (c, smtValueUnsafeCastInt (unsafeLookup c res')) | c <- z3Vars_coef ]
      in return (alg, prec)
  where
    esWithIds = [("__rulevar" ++ show i, e) | (i, e) <- zip ([0..] :: [Int]) (es ++ PES.inverse es)]
    -- NOTE: include goal (reduction orders require total precedence)
    fs = functionsInESWithArity (goal : es)
    reducibilityOnTerm t =
      orFormula [ BoolVar ruleId | (ruleId, (l, _)) <- esWithIds, encompass t l ]
    test_terms = nub [ u | (s, t) <- goal : es, u <- [s, t] ]
    maximizedConstraints = [ reducibilityOnTerm t | t <- test_terms ]
    -- constraits to be satisfied
    orientation = andFormula [ impliesFormula (BoolVar ruleId) (encodeWPO N l r)
                             | (ruleId, (l, r)) <- esWithIds ]
    precTotality = Distinct [Var f | f <- z3Vars_prec ] -- totality is required since precedence is presented as list
    coef1n_nonneg = andFormula [ Gt (Var v) (Val 0) | v <- z3Vars_coef1n ] -- strict monotonicity
    coef0_nat = andFormula [ Geq (Var v) (Val 0) | v <- z3Vars_coef0 ]
    -- variables
    z3Vars_coef1n = [ ith_coef f i | (f, n) <- fs, i <- [1..n] ]
    z3Vars_coef0 = [ ith_coef f 0 | (f, _n) <- fs ]
    z3Vars_coef = z3Vars_coef0 ++ z3Vars_coef1n
    z3Vars_prec = [ prec_name f | (f, _n) <- fs ]
    esIds = [ esId | (esId, _) <- esWithIds ]
    z3Input = maxsat (z3Vars_prec ++ z3Vars_coef)
                     esIds
                     maximizedConstraints
                     (andFormula [orientation, precTotality, coef1n_nonneg, coef0_nat])

findLPOParamViaGWPO :: PES -> PE -> IO Precedence
findLPOParamViaGWPO es goal = do
  res <- z3 "toma" z3Input
  case res of
    Nothing -> error "findLPOParamViaGWPO: no parameter is found." -- this error means bug?
    Just res' ->
      let prec = assoc2Prec [ (f, n) | f <- z3Vars_prec, SMTIntValue n <- [unsafeLookup f res'] ]
      in return prec
  where
    esWithIds = [("__rulevar" ++ show i, e) | (i, e) <- zip ([0..] :: [Int]) (es ++ PES.inverse es)]
    -- NOTE: include goal (reduction orders require total precedence)
    fs = functionsInESWithArity (goal : es)
    reducibilityOnTerm t =
      orFormula [ BoolVar ruleId | (ruleId, (l, _)) <- esWithIds, encompass t l ]
    test_terms = nub [ u | (s, t) <- goal : es, u <- [s, t] ]
    maximizedConstraints = [ reducibilityOnTerm t | t <- test_terms ]
    -- constraits to be satisfied
    orientation = andFormula [ impliesFormula (BoolVar ruleId) (encodeWPO Max l r)
                             | (ruleId, (l, r)) <- esWithIds ]
    precTotality = Distinct [Var f | f <- z3Vars_prec ] -- totality is required since precedence is presented as list
    -- variables
    z3Vars_prec = [ prec_name f | (f, _n) <- fs ]
    esIds = [ esId | (esId, _) <- esWithIds ]
    z3Input = maxsat z3Vars_prec
                     esIds
                     maximizedConstraints
                     (andFormula [orientation, precTotality])


-- NOTE: maximize the # of reducibility.
-- More precisely, maximizes the # of terms t
-- that satisfies the follwoing condition:
-- there exists l = r \in E \cup E^{-1} such that
-- s encompasses l and l >_lpo r.

-- NOTE:
-- Maximizing naively the # of terms t such that t is reducible in (E, >_lpo)
-- seems to generate too many smt variables (?)
-- Encoding itself was also too expensive (?)
-- As a result, the overall procedure became too slow (?)
-- I haven't investigated in detail though...

findReductionOrderParam :: Config -> PES -> PE -> IO ReductionOrderParam
findReductionOrderParam config es goal =
  case (order_class config) of
    LPO -> do prec <- findLPOPrec es goal
              return (LPOParam prec)
    GWPO -> do (alg, prec) <- Prover.findGWPOParam01 es goal
               return (GWPOParam prec alg)
    GWPO01 -> do (alg, prec) <- Prover.findGWPOParam01 es goal
                 return (GWPOParam prec alg)
    GWPON -> do (alg, prec) <- Prover.findGWPOParamN es goal
                return (GWPOParam prec alg)
    WPO -> do (alg, prec) <- Prover.findWPOParam es goal
              return (WPOParam prec alg)
    LPOviaGWPO -> do prec <- Prover.findLPOParamViaGWPO es goal
                     return (LPOParam prec)

generateLemmata :: Int -> ES -> ReductionOrderParam -> (Int, ES)
generateLemmata next_id es param = (next_id + length ret, ret)
  where cps = selectLemmata [ c | c@((s, t), _d) <- cp (param2ord param) es,
                                  not (ReductionOrder.groundJoinable param es s t),
                                  not (inES (s, t) es) ]
        ret = [ E { eqn = e, eqn_id = i, active = True, eqn_derivation = d}
              | (i, (e, d)) <- zip [next_id..] cps ]

prove' :: Config -> Int -> PE -> PES -> ES -> ES -> IO Proof
prove' conf next_id goal@(s, t) initES es deleted_acc = do
  param <- findReductionOrderParam conf (toPES es) goal
  let (next_id', es', deleted) = if inter_reduction conf
                                  then interreduce next_id param es
                                  else (next_id, es, [])
  let next_deleted = deleted ++ deleted_acc
  let gt = param2ord param
  let (rw_s, nf_s) = nf es' gt s
  let (rw_t, nf_t) = nf es' gt t
  let (next_id'', lemmata) = generateLemmata next_id' es' param
  if verbose conf
    then do
      putStrLn "reduction order:"
      putStrLn (show param)
      putStrLn "ES (simplified):"
      putStrLn (showOTRS gt es')
      putStrLn "critical pairs:"
      putStrLn (showPES (toPES lemmata))
    else return ()
  -- First, the prover checks if s and t are joinable.
  if nf_s == nf_t
    then let
           d = Goal { rw_l = rw_s, rw_r = rw_t }
           g' = E { eqn = goal, eqn_id = next_id',
                    active = True, eqn_derivation = d }
          in return (Join { proof_goal = g', proof_es = es', proof_deleted_es =  next_deleted })
    else if all (\(s'', t'') -> ReductionOrder.groundJoinable param es' s'' t'') initES && -- ground equivalence and
            ReductionOrder.groundConfluent param es' -- ground confluence
          then return (Complete { proof_es = es', proof_reduction_order_param = param, proof_deleted_es = next_deleted })
          else do let nextES = (es' ++ lemmata)
                  prove' conf next_id'' goal initES nextES next_deleted

data Config = Config { inter_reduction :: Bool, order_class :: ReductionOrder.Class, verbose :: Bool }

-- goal must be ground 
prove :: Config -> PE -> ES -> IO Proof
prove conf goal es = prove' conf (1 + maxId es) goal (toPES es) es []
