module Term where

import Data.List

data Term = V String
          | F String [Term]
          deriving Eq

type TermPair = (Term, Term)

-- Let gt :: TermOrder. 
-- gt s t = True means s > t.
type TermOrder = Term -> Term -> Bool

-- smart constructor
constant :: String -> Term
constant f = F f []

size :: Term -> Int
size (V _) = 1
size (F _ ts) = 1 + sum [ size t | t <- ts ]

type Position = [Int]

emptyPosition :: Position
emptyPosition = []

instance Show Term where
  show (V x)    = x
  show (F c []) = c
  show (F f ts) = f ++ "(" ++ intercalate "," [ show t | t <- ts ] ++ ")"

-- Var(t)
variables :: Term -> [String]
variables (V x)    = [x]
variables (F _f ts) = nub [ x | ti <- ts, x <- variables ti ]

isGround :: Term -> Bool
isGround t = variables t == []

varsInPair :: TermPair -> [String]
varsInPair (t, s) = nub (variables t ++ variables s)

functions :: Term -> [String]
functions (V _) = []
functions (F f ts) = nub (f : [f' | t <- ts, f' <- functions t])

functionsInPair :: TermPair -> [String]
functionsInPair (t, s) = nub (functions t ++ functions s)

-- NOTE: this function assumes that arity is consistent.
functionsWithArity :: Term -> [(String, Int)]
functionsWithArity (V _) = []
functionsWithArity (F f ts) =
  nub ((f, length ts) : [ f_arity | t <- ts, f_arity <- functionsWithArity t])

-- Pos(t)
positions :: Term -> [Position]
positions (V _x) = [[]]
positions (F _f ts) = ([]: [ i : p | (i, t) <- zip [0..] ts, p <- positions t])

functionPositions :: Term -> [Position]
functionPositions (V _x) = []
functionPositions (F _f ts) = ([]: [ i : p | (i, t) <- zip [0..] ts, p <- functionPositions t])

-- subtermAt t p = t|_p
subtermAt :: Term -> Position -> Term
subtermAt t [] = t
subtermAt (V _) _ = undefined
subtermAt (F _f ts) (i : p)
  | i < length ts = subtermAt (ts !! i) p
  | otherwise = undefined

subterms :: Term -> [Term]
subterms t = [subtermAt t p | p <- positions t]

strictSubterms :: Term -> [Term]
strictSubterms t = [subtermAt t p | p <- positions t, p /= []]

isSubtermOf :: Term -> Term -> Bool
isSubtermOf t1 t2 = t1 `elem` (subterms t2)

isStrictSubtermOf :: Term -> Term -> Bool
isStrictSubtermOf t1 t2 = t1 `elem` (strictSubterms t2)

-- replace t u p = t[u]_p
replace :: Term -> Term -> Position -> Term
replace _t u [] = u
replace (V _) _ _ = undefined
replace (F f ts) u (i : p)
  | i < length ts = F f ts'
  | otherwise = undefined
      where ts' = [ if i == j then (replace tj u p) else tj | (j, tj) <- zip [0..] ts]

-- [a,b,c] means a > b > c
type Precedence = [String]

showPrec :: [String] -> String
showPrec prec = intercalate " > " prec 

-- example: [("a", 1), ("b", 0), ("c", 2)] becomes ["a", "b", "c"]
assoc2Prec :: [(String, Int)] -> Precedence
assoc2Prec as = map fst (reverse (sortOn snd as))

-- `gtPrec > f g` tests `f > g`
gtPrec :: Precedence -> String -> String -> Bool
gtPrec [] f g = error (f ++ " and " ++ g ++ " is not found.")
gtPrec (h : hs) f g
  | f == g = False
  | h == f = True
  | h == g = False
  | otherwise = gtPrec hs f g

-- only linear polynomials are considered
type Algebra = [(String, Int)]
